/******************************************************************************/
/* PDirectory implementation file.                                            */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#define _BUILDING_POLYKIT_LIBRARY_

// PolyKit headers
#include "PDebug.h"
#include "PException.h"
#include "PDirectory.h"
#include "PList.h"
#include "GlobalFuncs.h"


/******************************************************************************/
/* PDirectory class                                                           */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/******************************************************************************/
PDirectory::PDirectory(void)
{
#if __p_os == __p_beos
	enumDir = NULL;

#elif __p_os == __p_windows
	
	// Initialize member variables
	searchHandle = INVALID_HANDLE_VALUE;
	enumInitialized = false;

#endif
}



/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "name" is the name of the directory to use.                        */
/******************************************************************************/
PDirectory::PDirectory(PString name)
{
	SetDirectory(name);

#if __p_os == __p_beos
	enumDir = NULL;

#elif __p_os == __p_windows

	// Initialize member variables
	searchHandle = INVALID_HANDLE_VALUE;
	enumInitialized = false;

#endif
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PDirectory::~PDirectory(void)
{
	EndEnum();
}



/******************************************************************************/
/* SetDirectory() will change the directory to the name given.                */
/*                                                                            */
/* Input:  "name" is the name of the new directory to use.                    */
/******************************************************************************/
void PDirectory::SetDirectory(PString name)
{
	// Make sure that the name ends with a slash
	if (name.Right(1) != P_DIRSLASH)
		name += P_DIRSLASH;

	dirName = name;
}



/******************************************************************************/
/* GetDirectory() will return the current directory name.                     */
/*                                                                            */
/* Output: Is the current directory name.                                     */
/******************************************************************************/
PString PDirectory::GetDirectory(void) const
{
	return (dirName);
}



/******************************************************************************/
/* Append() will append some extra directory names to the current name.       */
/*                                                                            */
/* Input:  "name" is the name of the extra directories to append.             */
/******************************************************************************/
void PDirectory::Append(PString name)
{
	// Make sure that the name ends with a slash
	if (name.Right(1) != P_DIRSLASH)
		name += P_DIRSLASH;

	// Remove a start slash if any
	if (name.Left(1) == P_DIRSLASH)
		name = name.Mid(1);

	dirName += name;
}



/******************************************************************************/
/* FindDirectory() will set the directory to the directory given.             */
/*                                                                            */
/* Input:  "type" is the type of directory you want to set the class to.      */
/*                                                                            */
/* Except: PFileException, PSystemException.                                  */
/******************************************************************************/
void PDirectory::FindDirectory(PFindType type)
{
#if __p_os == __p_beos

	BPath path;
	status_t retVal;

	// Check to see if it's the launch directory, because
	// it's a special case
	if (type == pLaunch)
	{
		app_info info;
		BEntry appEntry;

		// Find the directory
		if ((retVal = be_app->GetAppInfo(&info)) != B_OK)
			throw PFileException(ConvertOSError(retVal), "<< FindDirectory >>");

		if ((retVal = appEntry.SetTo(&info.ref)) != B_OK)
			throw PFileException(ConvertOSError(retVal), "<< FindDirectory >>");

		if ((retVal = appEntry.GetPath(&path)) != B_OK)
			throw PFileException(ConvertOSError(retVal), "<< FindDirectory >>");

		if ((retVal = path.GetParent(&path)) != B_OK)
			throw PFileException(ConvertOSError(retVal), "<< FindDirectory >>");

		// And set it
		SetDirectory(path.Path());
	}
	else
	{
		// All other types
		directory_which which;

		// Convert the type to BeOS types
		switch (type)
		{
			case pSettings:
				which = B_USER_SETTINGS_DIRECTORY;
				break;

			default:
				// Illegal type
				ASSERT(false);
				which = B_USER_DIRECTORY;
				break;
		}

		// Find the directory
		if ((retVal = find_directory(which, &path)) != B_OK)
			throw PFileException(ConvertOSError(retVal), "<< FindDirectory >>");

		// And set it
		SetDirectory(path.Path());
	}

#elif __p_os == __p_windows

	PString path;

	// Check if the specified type is the launch directory
	switch (type)
	{
		// Launch directory
		case pLaunch:
			char	fileName[MAX_PATH];
			int32	pos;

			// Get the file path for the file used to create the calling process
			if (!GetModuleFileName(NULL, fileName, MAX_PATH))
				throw PFileException(ConvertOSError(GetLastError()), "<< FindDirectory >>");

			// Set the directory path to this file path
			path = fileName;

			// Remove the file name from the directory path
			if ((pos = path.ReverseFind(P_DIRSLASH_C)) == -1)
				throw PFileException(ERROR_INVALID_NAME, path);
			path = path.Left(pos + 1);
			break;

		// Settings directory
		case pSettings:

			char buffer[MAX_PATH + 1];
			UINT res;

			// The settings directory is depending on the kind of Windows version
			switch (GetOSVersion())
			{
				// Windows 9x
				case pWindows95:
				case pWindows98:

					// Use the  directory as settings directory
					res = GetWindowsDirectory(buffer, MAX_PATH + 1);
					break;

				// Windows NT/2000
				case pWindowsNT:
				case pWindows2000:

					// Use the user profile directory as settings directory
					res = GetEnvironmentVariable("USERPROFILE", buffer, MAX_PATH + 1);
					break;

				// Unsupported OS
				default:
					ASSERT(false);
			}

			// Check for insufficient buffer space
			if (res > MAX_PATH + 1)
				throw PSystemException(P_GEN_ERR_INSUFFICIENT_BUFFER);

			// Check if some other error occured
			if (res == 0)
				throw PFileException(ConvertOSError(GetLastError()), path);

			path = buffer;
			break;
	}

	// Set directory to the found directory path
	SetDirectory(path);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* CreateDirectory() will create all the directories the object points to.    */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PDirectory::CreateDirectory(void) const
{
#if __p_os == __p_beos

	status_t retVal;

	if ((retVal = create_directory(dirName, 0755)) != B_OK)
		throw PFileException(ConvertOSError(retVal), dirName);

#elif __p_os == __p_windows

	BOOL dirCreated;

	// Attempt to create the directory
	dirCreated = ::CreateDirectory(dirName, NULL);

	// Check if the directory could not be created
	if (!dirCreated)
	{
		DWORD error;
		PString tempPath;
		int pos;
		PList<PString> dirNameList;

		//
		// Create all missing subdirectories in the directory path one at a time
		//

		// Copy the directory path name
		tempPath = dirName;

		//
		// Cut the directory path name into subdirectory names and
		// store these in a list until a subdirectory can be created.
		//
		do
		{
			// Get the last error
			error = GetLastError();

			// Exit if the directory already exits
			if (error == ERROR_ALREADY_EXISTS)
				return;

			// Else check if the directory path was not found
			else if (error == ERROR_PATH_NOT_FOUND)
			{
				// Throw away the last slash in the directory path
				tempPath = tempPath.Left(tempPath.GetLength() - 1);

				// Add the directory name in the tail of the directory path to the list
				pos = tempPath.ReverseFind(P_DIRSLASH_C);
				dirNameList.AddHead(tempPath.Mid(pos));

				// Cut away the directory name in the tail of the directory path
				tempPath = tempPath.Left(pos + 1);

				// Attempt to create the subdirectory at the current directory path name
				dirCreated = ::CreateDirectory(tempPath, NULL);
			}
			// Else throw a file exception
			else
				throw PFileException(ConvertOSError(error), dirName);
		}
		while (!dirCreated);

		//
		// Create all the missing subdirectories out from the list of directory names
		//

		// Throw away the last slash in the directory path
		tempPath = tempPath.Left(tempPath.GetLength() - 1);

		// Create all missing subdirectories one at a time
		for (int32 i = 0; i < dirNameList.CountItems(); i++)
		{
			// Add the next directory name in the list to the direcory path
			tempPath += dirNameList.GetItem(i);

			// Attempt to create the subdirectory
			dirCreated = ::CreateDirectory(tempPath, NULL);

			// If the subdirectory could not be created then throw a file exception
			if (!dirCreated)
				throw PFileException(ConvertOSError(error), dirName);
		}
	}

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Contains() will search the directory to see if the entry exists in it.     */
/*                                                                            */
/* Input:  "name" is the name of the entry.                                   */
/*         "type" is the type of the entry to search for.                     */
/*                                                                            */
/* Output: True if the entry exists, else false.                              */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
bool PDirectory::Contains(PString name, PEntryType type) const
{
#if __p_os == __p_beos

	BDirectory dir;
	int32 nodeFlags;
	status_t retVal;

	if ((retVal = dir.SetTo(dirName)) != B_OK)
		throw PFileException(ConvertOSError(retVal), dirName);

	// Convert the type to BeOS types
	switch (type)
	{
		case pFile:
			nodeFlags = B_FILE_NODE;
			break;

		case pDirectory:
			nodeFlags = B_DIRECTORY_NODE;
			break;

		case pLink:
			nodeFlags = B_SYMLINK_NODE;
			break;

		default:
			nodeFlags = B_ANY_NODE;
			break;
	}

	return (dir.Contains(name, nodeFlags));

#elif __p_os == __p_windows
	
	PDirectory dir;
	PString retName;
	PEntryType retType;

	// Set the directory name
	dir.SetDirectory(dirName);

	// Initilalize enumeration on the specified type
	dir.InitEnum(type);

	// Search for the specified name and type in the directories entries
	while (dir.GetNextEntry(retName, retType))
	{
		// Check if the specified type is a link
		if (retType == pLink)
		{
			// Construct the link name
			PString linkName = name;
			name += ".lnk";

			// Check if the link name
			if (retName.CompareNoCase(linkName) == 0)
			{
				// If so, end enumeration
				dir.EndEnum();
				return (true);
			}
		}

		// Check if the name of an entry match the specified name
		if (retName.CompareNoCase(name) == 0)
		{
			// If so, end enumeration
			dir.EndEnum();
			return (true);
		}
	}

	// End enumeration
	dir.EndEnum();
	return (false);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* InitEnum() will initialize the enumeration.                                */
/*                                                                            */
/* Input:  "type" is the type of the entry to enumerate.                      */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PDirectory::InitEnum(PEntryType type)
{
#if __p_os == __p_beos

	status_t retVal;

	// Forgot to call EndEnum()?
	ASSERT(enumDir == NULL);

	// Initialize the class variables
	enumType = type;
	enumDir  = new BDirectory;
	if (enumDir == NULL)
		throw PMemoryException();

	// Set the work directory
	if ((retVal = enumDir->SetTo(dirName)) != B_OK)
		throw PFileException(ConvertOSError(retVal), dirName);

	// Initialize the directory pointer
	if ((retVal = enumDir->Rewind()) != B_OK)
		throw PFileException(ConvertOSError(retVal), dirName);

#elif __p_os == __p_windows

	// Forgot to call EndEnum()?
	ASSERT(searchHandle == INVALID_HANDLE_VALUE);

	// Initialize the class variables
	enumType = type;
	enumInitialized = true;

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* GetNextEntry() will return the name of the next entry.                     */
/*                                                                            */
/* Input:  "name" is a reference to store the name of the entry.              */
/*         "type" is a reference to store the type of the entry.              */
/*                                                                            */
/* Output: True if an entry has been returned, else false.                    */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
bool PDirectory::GetNextEntry(PString &name, PEntryType &type)
{
#if __p_os == __p_beos

	BEntry entry;
	BPath path;
	status_t retVal;
	bool gotIt = false;

	do
	{
		// Get the next entry
		if ((retVal = enumDir->GetNextEntry(&entry)) != B_OK)
		{
			if (retVal == B_ENTRY_NOT_FOUND)
				return (false);

			throw PFileException(ConvertOSError(retVal), dirName);
		}

		// Check the type to see if we want to return it
		switch (enumType)
		{
			case pFile:
				if (entry.IsFile())
				{
					type  = pFile;
					gotIt = true;
				}
				break;

			case pDirectory:
				if (entry.IsDirectory())
				{
					type  = pDirectory;
					gotIt = true;
				}
				break;

			case pLink:
				if (entry.IsSymLink())
				{
					type  = pLink;
					gotIt = true;
				}
				break;

			default:
				if (entry.IsFile())
					type = pFile;
				else
				{
					if (entry.IsDirectory())
						type = pDirectory;
					else
					{
						if (entry.IsSymLink())
							type = pLink;
						else
							type = pAny;
					}
				}

				gotIt = true;
				break;
		}
	}
	while (!gotIt);

	// Okay, now get the name
	if ((retVal = entry.GetPath(&path)) != B_OK)
		throw PFileException(ConvertOSError(retVal), dirName);

	name = path.Leaf();
	return (true);

#elif __p_os == __p_windows

	bool gotIt = false;

	// Go through the files in the directory until the right entry type is found
	do
	{
		// Check if enumeration was just initialized. If so, FindFirstFile() must be used.
		if (enumInitialized)
		{
			// Find the first file in the directory
			PString anyFileName = dirName;
			anyFileName += "*";
			searchHandle = FindFirstFile(anyFileName, &findData);

			// Check if the search handle is invalid
			if (searchHandle == INVALID_HANDLE_VALUE)
			{
				// Get the last error
				DWORD error = GetLastError();

				// Check if the error is cause because no files exists in the directory
				if (error == ERROR_PATH_NOT_FOUND)
					return (false);

				throw PFileException(ConvertOSError(error), dirName);
			}

			// When this function has finished, the enumberation initializing is over
			enumInitialized = false;
		}
		// Else FindNextFile() must be used instead
		else
		{
			// Attempt to find the next file in the directory
			if (FindNextFile(searchHandle, &findData) == 0)
			{
				// Get the last error
				DWORD error = GetLastError();

				// Check if the error is caused by no more files in the directory
				if (error == ERROR_NO_MORE_FILES)
					return (false);

				// Else throw a file exception
				throw PFileException(ConvertOSError(error), dirName);
			}

			// Assign the entry name
			name = findData.cFileName;

			// Check if the entry name is dots (is excluded)
			if ((name != ".") && (name != ".."))
			{
				// Get the entry type for the current search handle

				// Check if the entry type is a directory
				if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
					type = pDirectory;

				// Else..
				else
				{
					// Check if the entry type is a link
					PString fileName = findData.cFileName;
					if (fileName.Right(4).CompareNoCase(".lnk") == 0)
						type = pLink;

					// Else the type is a ordinary file
					else
						type = pFile;
				}

				// Check if we got the specified entry type
				if ((enumType == type) || (enumType == pAny))
					gotIt = true;
			}
		}
	}
	while (!gotIt);

	return (true);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* EndEnum() will stop the enumeration and cleanup.                           */
/******************************************************************************/
void PDirectory::EndEnum(void)
{
#if __p_os == __p_beos

	// Delete the BDirectory object
	delete enumDir;
	enumDir = NULL;

#elif __p_os == __p_windows

	// Close the search handle from a enumeration
	if (searchHandle != INVALID_HANDLE_VALUE)
		VERIFY(FindClose(searchHandle));
	searchHandle = INVALID_HANDLE_VALUE;

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* GetDirectoryPart() will return the directory part of a path.               */
/*                                                                            */
/* Input:  "path" is the path to return the directory part of.                */
/*                                                                            */
/* Output: The directory part.                                                */
/******************************************************************************/
PString PDirectory::GetDirectoryPart(PString path)
{
	int32 index;

	// Find the last slash
	index = path.ReverseFind(P_DIRSLASH_C);
	if (index == -1)
		return ("");

	return (path.Left(index + 1));
}



/******************************************************************************/
/* GetFilePart() will return the file part of a path.                         */
/*                                                                            */
/* Input:  "path" is the path to return the file part of.                     */
/*                                                                            */
/* Output: The file part.                                                     */
/******************************************************************************/
PString PDirectory::GetFilePart(PString path)
{
	int32 index;

	// Check if the path is a directory only
	if (path.Right(1) == P_DIRSLASH)
		return ("");

	// Find the last slash
	index = path.ReverseFind(P_DIRSLASH_C);
	if (index == -1)
		return (path);

	return (path.Mid(index + 1));
}
