/******************************************************************************/
/* PRegistry implementation file.                                             */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#define _BUILDING_POLYKIT_LIBRARY_

// PolyKit headers
#include "PDebug.h"
#include "PException.h"
#include "PString.h"
#include "PSynchronize.h"
#include "PFile.h"
#include "PDirectory.h"
#include "PRegistry.h"
#include "GlobalFuncs.h"


/******************************************************************************/
/* Default constructor                                                        */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
PRegistry::PRegistry(void)
{
	// Allocate objects needed
	lock = new PSWMRLocker;
	if (lock == NULL)
		throw PMemoryException();

	root = new PRegistryData;
	if (root == NULL)
	{
		delete lock;
		throw PMemoryException();
	}

	// Initialize the root structure
	root->next     = NULL;
	root->openCnt  = 0;
	root->type     = pKey;
	root->data     = NULL;
	root->dataSize = 0;

	changed   = false;
	dataBlock = NULL;
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PRegistry::~PRegistry(void)
{
	// Delete the datablock if any
	delete[] dataBlock;

	// Clean up the registry
	DeleteKey(root);

	// Delete the objects
	delete root;
	delete lock;
}



/******************************************************************************/
/* OpenKey() opens a key with the path given. You have to call this function  */
/*      before you can use any of the other functions.                        */
/*                                                                            */
/* Input:  "path" is the path to the key you want to open.                    */
/*                                                                            */
/* Output: A handle to use with the other functions.                          */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void *PRegistry::OpenKey(PString path)
{
	int32 prevPos = -1;
	int32 pos;
	PString tempStr;
	PRegistryData tempStruct;
	PRegistryData *handle;

	// The first character has to be a slash
	ASSERT(path.GetAt(0) == '/');

	// The last character has to be a slash
	if (path.GetAt(path.GetByteLength() - 1) != '/')
		path += "/";

	// Lock the list
	if (lock->WaitToWrite() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		// Initialize list pointer
		tempStruct.data = (void *)root;
		handle = &tempStruct;

		do
		{
			// Clip the first part of the string out
			pos = path.Find('/', prevPos);
			tempStr = path.Mid(prevPos + 1, pos - (prevPos + 1));
			prevPos = pos;

			// Go through the list to see if we can find the 
			handle = (PRegistryData *)handle->data;
			if (handle == NULL)
				throw PFileException(P_FILE_ERR_ENTRY_NOT_FOUND, path);

			do
			{
				// Did we found it yet
				if ((tempStr == handle->name) && (handle->type == pKey))
					break;

				// Next item in the list
				handle = handle->next;
			}
			while (handle != NULL);

			if (handle == NULL)
				throw PFileException(P_FILE_ERR_ENTRY_NOT_FOUND, path);
		}
		while (pos + 1 < path.GetByteLength());

		// Found it, now count up the open count
		AtomicIncrement((int32 *)&handle->openCnt);

		// Unlock it again
		lock->DoneWriting();
	}
	catch(...)
	{
		lock->DoneWriting();
		throw;
	}

	return (handle);
}



/******************************************************************************/
/* CloseKey() closes the key you have opened with the OpenKey() function.     */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::CloseKey(void *handle)
{
	if (handle)
	{
		PRegistryData *hd = (PRegistryData *)handle;

		// Lock the list
		if (lock->WaitToWrite() != pSyncOk)
			throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

		ASSERT(hd->openCnt != 0);

		// Count down the open count
		AtomicDecrement((int32 *)&hd->openCnt);

		// And unlock it again
		lock->DoneWriting();
	}
}



/******************************************************************************/
/* Exists() will check to see if the key or value is created in the opened    */
/*         key.                                                               */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the key or value you want to search for.     */
/*                                                                            */
/* Output: True if it exists, else false.                                     */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
bool PRegistry::Exists(void *handle, PString name)
{
	PRegistryData *curEntry;

	// Lock the list
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	curEntry = (PRegistryData *)((PRegistryData *)handle)->data;

	while (curEntry != NULL)
	{
		if (curEntry->name == name)
		{
			// And unlock it again
			lock->DoneReading();
			return (true);
		}

		curEntry = curEntry->next;
	}

	// And unlock it again
	lock->DoneReading();
	return (false);
}



/******************************************************************************/
/* CreateKey() will create a new key in the opened key.                       */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the key you want to create.                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::CreateKey(void *handle, PString name)
{
	PRegistryData *lastEntry, *curEntry;

	// First check to see if the key already exists
	if (!(Exists(handle, name)))
	{
		// It doesn't, so find the last entry in the opened key
		lastEntry = FindLast((PRegistryData *)handle);

		// Now allocate the new entry
		curEntry = new PRegistryData;
		if (curEntry == NULL)
			throw PMemoryException();

		// Initialize the new allocated key
		curEntry->next     = NULL;
		curEntry->openCnt  = 0;
		curEntry->type     = pKey;
		curEntry->name     = name;
		curEntry->data     = NULL;
		curEntry->dataSize = 0;

		// We want unique access to the list
		if (lock->WaitToWrite() != pSyncOk)
		{
			delete curEntry;
			throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");
		}

		// Link the entry to the list
		lastEntry->next = curEntry;

		// The registry has been changed
		changed = true;

		// Unlock again
		lock->DoneWriting();
	}
}



/******************************************************************************/
/* CreateData() will create a new value in the opened key with a block of     */
/*         data.                                                              */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the value you want to create.                */
/*         "dataAdr" is a pointer to the data you want stored in the value.   */
/*         "size" is the size of the data block in bytes.                     */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::CreateData(void *handle, PString name, const void *dataAdr, uint32 size)
{
	CreateValue((PRegistryData *)handle, name, pData, dataAdr, size);
}



/******************************************************************************/
/* CreateNumber() will create a new value in the opened key with a number.    */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the value you want to create.                */
/*         "num" is the number you want to store.                             */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::CreateNumber(void *handle, PString name, int32 num)
{
	CreateValue((PRegistryData *)handle, name, pNumber, &num, sizeof(uint32));
}



/******************************************************************************/
/* CreateString() will create a new value in the opened key with a string.    */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the value you want to create.                */
/*         "str" is the string you want to store.                             */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::CreateString(void *handle, PString name, PString str)
{
	CreateValue((PRegistryData *)handle, name, pString, str, str.GetByteLength() + 1);
}



/******************************************************************************/
/* RemoveKey() will remove a key from the opened key and all its entries.     */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the key you want to remove.                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::RemoveKey(void *handle, PString name)
{
	PRegistryData *entry, *prev;

	// We want unique access to the list
	if (lock->WaitToWrite() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		entry = FindEntry((PRegistryData *)handle, name, &prev);
		if (entry == NULL)
			throw PFileException(P_FILE_ERR_ENTRY_NOT_FOUND, name);

		// Check to see if anybody else has the key opened
		if (entry->openCnt > 0)
			throw PFileException(P_GEN_ERR_CANT_PERFORM, name);

		// Unlink the key we want to remove
		prev->next = entry->next;

		// Remove the key from the memory
		DeleteKey(entry);

		// Remove the key from the memory
		delete entry;

		// The registry has been changed
		changed = true;

		// Unlock it again
		lock->DoneWriting();
	}
	catch(...)
	{
		lock->DoneWriting();
		throw;
	}
}



/******************************************************************************/
/* RemoveKeyEntries() will remove all the entries from a key in the opened    */
/*         key.                                                               */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the key you want to remove the entries from. */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::RemoveKeyEntries(void *handle, PString name)
{
	PRegistryData *entry, *prev;

	// We want unique access to the list
	if (lock->WaitToWrite() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		entry = FindEntry((PRegistryData *)handle, name, &prev);
		if (entry == NULL)
			throw PFileException(P_FILE_ERR_ENTRY_NOT_FOUND, name);

		// Remove the key entries from the memory
		DeleteKey(entry);
		entry->data     = NULL;
		entry->dataSize = 0;

		// The registry has been changed
		changed = true;

		// Unlock it again
		lock->DoneWriting();
	}
	catch(...)
	{
		lock->DoneWriting();
		throw;
	}
}



/******************************************************************************/
/* RemoveValue() will remove a value from the opened key.                     */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the value you want to remove.                */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::RemoveValue(void *handle, PString name)
{
	PRegistryData *entry, *prev;

	// We want unique access to the list
	if (lock->WaitToWrite() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		entry = FindEntry((PRegistryData *)handle, name, &prev);
		if (entry == NULL)
			throw PFileException(P_FILE_ERR_ENTRY_NOT_FOUND, name);

		// Unlink the value we want to remove
		prev->next = entry->next;

		// Remove the value from the memory
		delete entry->data;
		delete entry;

		// The registry has been changed
		changed = true;

		// Unlock it again
		lock->DoneWriting();
	}
	catch(...)
	{
		lock->DoneWriting();
		throw;
	}
}



/******************************************************************************/
/* GetData() will return the data from value in the opened key.               */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the value you want to get the data from.     */
/*         "size" is a pointer to where to store the size of the data block.  */
/*                                                                            */
/* Output: A pointer to the value data.                                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void *PRegistry::GetData(void *handle, PString name, uint32 *size)
{
	void *data;

	// Get access to the data
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		// Get the data
		data = GetValue((PRegistryData *)handle, name, size);

		// Allocate a new memory block to hold the data
		delete[] dataBlock;
		dataBlock = NULL;

		dataBlock = new uint8[*size];
		if (dataBlock == NULL)
			throw PMemoryException();

		// Copy the data to the new memory block
		memcpy(dataBlock, data, *size);

		// Unlock again
		lock->DoneReading();
	}
	catch(...)
	{
		lock->DoneReading();
		throw;
	}

	return (dataBlock);
}



/******************************************************************************/
/* GetNumber() will return the number from value in the opened key.           */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the value you want to get the number from.   */
/*                                                                            */
/* Output: The number stored in the value.                                    */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PRegistry::GetNumber(void *handle, PString name)
{
	uint32 *dataAdr;
	uint32 size, num;

	// Get access to the data
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		// Get the data and size
		dataAdr = (uint32 *)GetValue((PRegistryData *)handle, name, &size);
		num = *dataAdr;

		// Unlock again
		lock->DoneReading();
	}
	catch(...)
	{
		lock->DoneReading();
		throw;
	}

	return (num);
}



/******************************************************************************/
/* GetString() will return the string from value in the opened key.           */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*         "name" is the name of the value you want to get the string from.   */
/*                                                                            */
/* Output: The string stored in the value.                                    */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PRegistry::GetString(void *handle, PString name)
{
	void *dataAdr;
	uint32 size;
	PString retStr;
	char *strBuffer;

	// Get access to the data
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		// Get the data and size
		dataAdr = GetValue((PRegistryData *)handle, name, &size);

		// Initialize string
		strBuffer = retStr.GetBufferSetLength(size);
		memcpy(strBuffer, dataAdr, size);
		retStr.ReleaseBuffer();

		// Unlock again
		lock->DoneReading();
	}
	catch(...)
	{
		lock->DoneReading();
		throw;
	}

	return (retStr);
}



/******************************************************************************/
/* InitValueEnum() initializes the enumeration in the current key after       */
/*         values.                                                            */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*                                                                            */
/* Output: A enumeration handle or NULL if the key is empty.                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void *PRegistry::InitValueEnum(void *handle)
{
	PRegistryData *entry;

	// Lock the list
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	entry = (PRegistryData *)((PRegistryData *)handle)->data;

	// Skip subkeys
	if (entry != NULL)
	{
		while (entry->type == pKey)
		{
			entry = entry->next;
			if (entry == NULL)
				break;
		}
	}

	// Unlock it again
	lock->DoneReading();
	return (entry);
}



/******************************************************************************/
/* GetNextValue() will return the next value in the enumeration. It will skip */
/*         the subkeys.                                                       */
/*                                                                            */
/* Input:  "handle" is the handle you got from the InitValueEnum() function.  */
/*         "name" is a reference to a string to store the value name.         */
/*         "buffer" is a pointer to store the pointer to the data.            */
/*         "size" is a pointer to where to store the size of the data block.  */
/*         "type" is a pointer to where to store the type of the data.        */
/*                                                                            */
/* Output: A new handle to give to the next function call or NULL if it's the */
/*         last entry in the key.                                             */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void *PRegistry::GetNextValue(void *handle, PString &name, void **buffer, uint32 *size, PRegKind *type)
{
	PRegistryData *entry;

	ASSERT(handle != NULL);
	ASSERT(buffer != NULL);
	ASSERT(size != NULL);
	ASSERT(type != NULL);

	// Lock the list
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	entry = (PRegistryData *)handle;

	// Copy the pointers and name
	*buffer = entry->data;
	*size   = entry->dataSize;
	*type   = entry->type;
	name    = entry->name;

	// Skip entry
	entry = entry->next;

	// Skip subkeys
	if (entry != NULL)
	{
		while (entry->type == pKey)
		{
			entry = entry->next;
			if (entry == NULL)
				break;
		}
	}

	// Unlock it again
	lock->DoneReading();
	return (entry);
}



/******************************************************************************/
/* InitKeyEnum() initializes the enumeration in the current key after keys.   */
/*                                                                            */
/* Input:  "handle" is the handle you got from the OpenKey() function.        */
/*                                                                            */
/* Output: A enumeration handle or NULL if the key is empty.                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void *PRegistry::InitKeyEnum(void *handle)
{
	PRegistryData *entry;

	// Lock the list
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	entry = (PRegistryData *)((PRegistryData *)handle)->data;

	// Skip values
	if (entry != NULL)
	{
		while (entry->type != pKey)
		{
			entry = entry->next;
			if (entry == NULL)
				break;
		}
	}

	// Unlock it again
	lock->DoneReading();
	return (entry);
}



/******************************************************************************/
/* GetNextKey() will return the next key in the enumeration. It will skip all */
/*         the values.                                                        */
/*                                                                            */
/* Input:  "handle" is the handle you got from the InitKeyEnum() function.    */
/*         "name" is a reference to a string to store the entry name.         */
/*                                                                            */
/* Output: A new handle to give to the next function call or NULL if it's the */
/*         last entry in the key.                                             */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void *PRegistry::GetNextKey(void *handle, PString &name)
{
	PRegistryData *entry;

	ASSERT(handle != NULL);

	// Lock the list
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	entry = (PRegistryData *)handle;

	// Copy the name
	name = entry->name;

	// Skip entry
	entry = entry->next;

	// Skip values
	if (entry != NULL)
	{
		while (entry->type != pKey)
		{
			entry = entry->next;
			if (entry == NULL)
				break;
		}
	}

	// Unlock it again
	lock->DoneReading();
	return (entry);
}



/******************************************************************************/
/* CopyKey() will copy all entries and subkeys from the source key into the   */
/*         destination key you give.                                          */
/*                                                                            */
/* Input:  "source" is the source key you want to copy from.                  */
/*         "destination" is the destination key you want to copy into.        */
/*         "destReg" is a pointer to the destination registry or NULL for     */
/*         the current objects registry.                                      */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::CopyKey(PString source, PString destination, PRegistry *destReg)
{
	void *sourceHandle, *destHandle, *enumHandle;
	PString name;
	void *buffer;
	uint32 size;
	PRegKind type;

	// The last character has to be a slash
	if (source.GetAt(source.GetByteLength() - 1) != '/')
		source += "/";

	if (destination.GetAt(destination.GetByteLength() - 1) != '/')
		destination += "/";

	if (destReg == NULL)
		destReg = this;

	// Start to open the keys
	sourceHandle = OpenKey(source);

	try
	{
		destHandle = destReg->OpenKey(destination);

		try
		{
			// Initialize enumeration
			enumHandle = InitValueEnum(sourceHandle);

			// Go throu all the entries
			while (enumHandle != NULL)
			{
				enumHandle = GetNextValue(enumHandle, name, &buffer, &size, &type);

				destReg->CreateValue((PRegistryData *)destHandle, name, type, buffer, size);
			}

			// Now take the keys
			enumHandle = InitKeyEnum(sourceHandle);

			while (enumHandle != NULL)
			{
				enumHandle = GetNextKey(enumHandle, name);

				destReg->CreateKey(destHandle, name);
				CopyKey(source + name + "/", destination + name + "/", destReg);
			}
		}
		catch(...)
		{
			destReg->CloseKey(destHandle);
			throw;
		}

		// Close the keys again
		destReg->CloseKey(destHandle);
	}
	catch(...)
	{
		CloseKey(sourceHandle);
		throw;
	}

	CloseKey(sourceHandle);
}



/******************************************************************************/
/* LoadReg() loads the registry from disk from the user settings directory.   */
/*                                                                            */
/* Input:  "company" is the name of the company.                              */
/*         "program" is the name of the program.                              */
/*         "fileName" is the file to load without path.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::LoadReg(PString company, PString program, PString fileName)
{
	PDirectory dir;
	PCacheFile file(1024);
	PString fullName;

	// We want unique access to the list
	if (lock->WaitToWrite() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		// First we have to delete the old registry in memory
		DeleteKey(root);

		// Initialize the root structure
		root->next     = NULL;
		root->openCnt  = 0;
		root->type     = pKey;
		root->data     = NULL;
		root->dataSize = 0;

		// Unlock again
		lock->DoneWriting();
	}
	catch(...)
	{
		lock->DoneWriting();
		throw;
	}

	// Create configuration file name
	dir.FindDirectory(PDirectory::pSettings);
	dir.Append(company);
	dir.Append(program);

	// Create directories if they doesn't exists
	dir.CreateDirectory();

	// Open the file
	fullName = dir.GetDirectory() + fileName;
	file.Open(fullName, PFile::pModeRead);

	if (file.GetLength() < 4)
		throw PFileException(P_GEN_ERR_BAD_IDENTIFIER, fullName);

	// Check the identifier
	if (file.Read_B_UINT32() != 'REG0')
		throw PFileException(P_GEN_ERR_BAD_IDENTIFIER, fullName);

	// Do only initialize the registry if there is any
	if (file.GetLength() > 4)
	{
		// Begin to put the data into the registry
		LoadKey(&file, 4, root);
	}

	// A new registry has been loaded
	changed = false;
}



/******************************************************************************/
/* SaveReg() saves the registry to disk in the user settings directory.       */
/*                                                                            */
/* Input:  "company" is the name of the company.                              */
/*         "program" is the name of the program.                              */
/*         "fileName" is the file to load without path.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::SaveReg(PString company, PString program, PString fileName)
{
	PDirectory dir;
	PFile file;
	PString fullName;

	// Lock the list
	if (lock->WaitToRead() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		// Do only save the registry, if it has been changed
		if (changed)
		{
			// Create configuration file name
			dir.FindDirectory(PDirectory::pSettings);
			dir.Append(company);
			dir.Append(program);

			// Open the file
			fullName = dir.GetDirectory() + fileName;
			file.Open(fullName, PFile::pModeReadWrite | PFile::pModeCreate);

			// Write file identifier
			file.Write_B_UINT32('REG0');

			// It's the registry not empty, we want to write it to the disk
			if (root->data)
			{
				// Save the whole registry
				SaveKey(&file, (PRegistryData *)root->data);
			}

#if __p_os == __p_beos

			// Finally, we set the mime type
			file.SetFileType("application/octet-stream");

#endif
		}

		// The registry has been saved
		changed = false;

		// Unlock again
		lock->DoneReading();
	}
	catch(...)
	{
		lock->DoneReading();
		throw;
	}
}



/******************************************************************************/
/* CloneReg() clones a registry object into the current object. The two       */
/*         objects will have the same entries after the cloning.              */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::CloneReg(PRegistry *source)
{
	// We want unique access to the list
	if (lock->WaitToWrite() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		// Clean up the registry
		DeleteKey(root);

		// Initialize the root structure
		root->next     = NULL;
		root->openCnt  = 0;
		root->type     = pKey;
		root->data     = NULL;
		root->dataSize = 0;

		// Unlock again
		lock->DoneWriting();
	}
	catch(...)
	{
		lock->DoneWriting();
		throw;
	}

	// Call recursive function which will copy one key and go into the subkeys
	CopyKey("/", source);

	// The registry has been changed, but we don't mark it!
	changed = false;
}



/******************************************************************************/
/* FindLast() will search through the list in the start key and return the    */
/*         last entry in the list.                                            */
/*                                                                            */
/* Input:  "start" is a pointer to the entry to start the search in.          */
/*                                                                            */
/* Output: A pointer to the last entry in the list.                           */
/******************************************************************************/
PRegistry::PRegistryData *PRegistry::FindLast(PRegistryData *start)
{
	if (start->data == NULL)
		return ((PRegistryData *)&start->data);

	// Go into the directory
	start = (PRegistryData *)start->data;

	while (start->next != NULL)
		start = start->next;

	return (start);
}



/******************************************************************************/
/* FindEntry() will search through the list in the start key and return the   */
/*         entry where it found the name or NULL if it couldn't be found.     */
/*                                                                            */
/* Input:  "handle" is a pointer to the entry to start the search in.         */
/*         "name" is the name of the key or value you want to search for.     */
/*         "prevEntry" is a pointer where to store the previous entry or null.*/
/*                                                                            */
/* Output: A pointer to the entry or NULL if it couldn't be found.            */
/******************************************************************************/
PRegistry::PRegistryData *PRegistry::FindEntry(PRegistryData *start, PString name, PRegistryData **prevEntry)
{
	PRegistryData *prev;

	prev  = (PRegistryData *)&start->data;
	start = (PRegistryData *)start->data;

	while (start != NULL)
	{
		if (start->name == name)
		{
			if (prevEntry != NULL)
				*prevEntry = prev;

			return (start);
		}

		// Go to the next entry
		prev  = start;
		start = start->next;
	}

	return (NULL);
}



/******************************************************************************/
/* CreateValue() will create a new value in the opened key with a block of    */
/*         data.                                                              */
/*                                                                            */
/* Input:  "workKey" is the key to work in.                                   */
/*         "name" is the name of the value you want to create.                */
/*         "type" is the type of the data.                                    */
/*         "dataAdr" is a pointer to the data you want stored in the value.   */
/*         "size" is the size of the data block in bytes.                     */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::CreateValue(PRegistryData *workKey, PString name, PRegKind type, const void *dataAdr, uint32 size)
{
	PRegistryData *curEntry, *prevEntry;

	ASSERT(type != pKey);
	ASSERT(dataAdr != NULL);
	ASSERT(size != 0);

	// We want unique access to the list
	if (lock->WaitToWrite() != pSyncOk)
		throw PFileException(P_GEN_ERR_CANT_PERFORM, "<< Locker >>");

	try
	{
		// First see if the value already exists
		curEntry = FindEntry(workKey, name);
		if (curEntry == NULL)
		{
			// Create a new entry
			curEntry = new PRegistryData;
			if (curEntry == NULL)
				throw PMemoryException();

			// Initialize the new allocated value
			curEntry->next     = NULL;
			curEntry->openCnt  = 0;
			curEntry->name     = name;

			// Find the last entry in the list
			prevEntry = FindLast(workKey);

			// Link the entry to the list
			prevEntry->next = curEntry;
		}
		else
		{
			// Value already exists, but it is a key?
			if (curEntry->type == pKey)
				throw PFileException(P_FILE_ERR_IS_A_DIRECTORY, name);

			// All things it's ok, so just change the data, but first delete
			// the old data block
			delete curEntry->data;
		}

		// Allocate the new block of data
		curEntry->data = new uint8[size];
		if (curEntry->data == NULL)
			throw PMemoryException();

		// Copy the data
		memcpy(curEntry->data, dataAdr, size);

		// Initialize the data part of the registry structure
		curEntry->dataSize = size;
		curEntry->type     = type;

		// The registry has been changed
		changed = true;

		// Unlock again
		lock->DoneWriting();
	}
	catch(...)
	{
		lock->DoneWriting();
		throw;
	}
}



/******************************************************************************/
/* DeleteKey() will delete all entries and subkeys in the key you give, but   */
/*         not the key itself.                                                */
/*                                                                            */
/* Input:  "curKey" is a pointer to key you want to delete the entries from.  */
/******************************************************************************/
void PRegistry::DeleteKey(PRegistryData *curKey)
{
	PRegistryData *workEntry;

	// Go into the directory
	curKey = (PRegistryData *)curKey->data;

	while (curKey != NULL)
	{
		workEntry = curKey;
		curKey    = curKey->next;

		// If the current entry is a key, we go into it
		if (workEntry->type == pKey)
			DeleteKey(workEntry);
		else
			delete[] workEntry->data;			// Delete the data block

		// Delete the key
		delete workEntry;
	}
}



/******************************************************************************/
/* GetValue() will return the pointer to the data in the opened key to the    */
/*         value given.                                                       */
/*                                                                            */
/* Input:  "workKey" is the key to work in.                                   */
/*         "name" is the name of the value you want to get the data from.     */
/*         "size" is a pointer to store the size of the data block.           */
/*                                                                            */
/* Output: Is the pointer to the data.                                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void *PRegistry::GetValue(PRegistryData *workKey, PString name, uint32 *size)
{
	PRegistryData *curEntry;

	// First see if the value already exists
	curEntry = FindEntry(workKey, name);
	if (curEntry == NULL)
		throw PFileException(P_FILE_ERR_ENTRY_NOT_FOUND, name);

	// Check to see if the found entry is a key
	if (curEntry->type == pKey)
		throw PFileException(P_FILE_ERR_IS_A_DIRECTORY, name);

	// Get the size of the data
	*size = curEntry->dataSize;

	// Return the data pointer
	return (curEntry->data);
}



/******************************************************************************/
/* LoadKey() will load all entries and subkeys into the key you give, but not */
/*         the key itself.                                                    */
/*                                                                            */
/* Input:  "file" is a pointer to the file to load the data from.             */
/*         "position" is where in the file the entry to load is stored.       */
/*         "curKey" is a pointer to key you want to load the entries into.    */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::LoadKey(PFile *file, int32 position, PRegistryData *curKey)
{
	PDiskData curEntry;
	PRegistryData *lastEntry;
	PString name;
	uint8 *buffer;

	while (position != 0)
	{
		// Seek to the entry position
		file->Seek(position, PFile::pSeekBegin);

		// Read the entry from the file
		curEntry.next     = file->Read_B_UINT32();
		curEntry.type     = file->Read_B_UINT32();
		curEntry.data     = file->Read_B_UINT32();
		curEntry.dataSize = file->Read_B_UINT32();
		curEntry.nameLen  = file->Read_B_UINT32();

		// Copy entry name into the PString
		buffer = (uint8 *)name.GetBufferSetLength(curEntry.nameLen);
		file->Read(buffer, curEntry.nameLen);
		name.ReleaseBuffer(curEntry.nameLen);

		if (curEntry.type == pKey)
		{
			// Create a new key a create all its entries
			CreateKey(curKey, name);

			// Find the last entry in the opened key
			lastEntry = FindLast(curKey);
			LoadKey(file, curEntry.data, lastEntry);
		}
		else
		{
			// Seek to the data
			file->Seek(curEntry.data, PFile::pSeekBegin);

			// Create a new value
			switch (curEntry.type)
			{
				//
				// Data and String
				//
				case pData:
				case pString:
					// Allocate buffer to hold the data
					buffer = new uint8[curEntry.dataSize];
					if (buffer == NULL)
						throw PMemoryException();

					try
					{
						// Load the data into the buffer
						file->Read(buffer, curEntry.dataSize);

						if (curEntry.type == pString)
						{
							// Create the string
							CreateString(curKey, name, (const char *)buffer);
						}
						else
						{
							// Create the data entry
							CreateData(curKey, name, buffer, curEntry.dataSize);
						}

						// Free the temp buffer
						delete[] buffer;
					}
					catch(...)
					{
						delete[] buffer;
						throw;
					}
					break;

				//
				// Number
				//
				case pNumber:
					CreateNumber(curKey, name, file->Read_B_UINT32());
					break;

				//
				// Unknown type
				//
				default:
					ASSERT(false);
			}
		}

		// Go to the next entry
		position = curEntry.next;
	}
}



/******************************************************************************/
/* SaveKey() will save all entries and subkeys in the key you give, but not   */
/*         the key itself.                                                    */
/*                                                                            */
/* Input:  "file" is a pointer to the file to save the data to.               */
/*         "curKey" is a pointer to key you want to save the entries from.    */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PRegistry::SaveKey(PFile *file, PRegistryData *curKey)
{
	int32 pos, dataPos, nextPos;
	int32 entryLen;

	while (curKey != NULL)
	{
		// Get the current file position
		pos      = file->GetPosition();
		entryLen = curKey->name.GetByteLength();

		// Write the entry to disk
		file->Write_B_UINT32(0);					// next
		file->Write_B_UINT32(curKey->type);			// type
		file->Write_B_UINT32(0);					// data
		file->Write_B_UINT32(curKey->dataSize);		// dataSize
		file->Write_B_UINT32(entryLen);				// nameLen

		// Write the entry name
		file->Write(curKey->name, entryLen);

		// Remember the position of the data
		dataPos = file->GetPosition();

		// Write the entry data
		switch (curKey->type)
		{
			//
			// Key
			//
			case pKey:
			{
				// If the current entry is a key, we go into it
				SaveKey(file, (PRegistryData *)curKey->data);
				break;
			}

			//
			// Number
			//
			case pNumber:
			{
				file->Write_B_UINT32(*((uint32 *)curKey->data));
				break;
			}

			//
			// All other types
			//
			default:
			{
				// Save the data
				file->Write(curKey->data, curKey->dataSize);
			}
		}

		// Get the position of the next entry
		nextPos = file->GetPosition();

		// Write the data position
		if (curKey->data != NULL)
		{
			file->Seek(pos + 8, PFile::pSeekBegin);
			file->Write_B_UINT32(dataPos);
		}

		// Write the next position
		if (curKey->next != NULL)
		{
			file->Seek(pos, PFile::pSeekBegin);
			file->Write_B_UINT32(nextPos);
		}

		// Seek to the next position
		file->Seek(nextPos, PFile::pSeekBegin);

		// Go to the next entry
		curKey = curKey->next;
	}
}



/******************************************************************************/
/* CopyKey() will copy all entries and subkeys in the key you give            */
/*                                                                            */
/* Input:  "key" is the key you want to copy.                                 */
/*         "source" is a pointer to a registry object to copy from.           */
/******************************************************************************/
void PRegistry::CopyKey(PString key, PRegistry *source)
{
	void *sourceHandle, *destHandle, *enumHandle;
	PString name;
	void *buffer;
	uint32 size;
	PRegKind type;

	// Start to open the keys
	sourceHandle = source->OpenKey(key);
	destHandle   = OpenKey(key);

	// Initialize enumeration
	enumHandle = source->InitValueEnum(sourceHandle);

	// Go throu all the entries
	while (enumHandle != NULL)
	{
		enumHandle = source->GetNextValue(enumHandle, name, &buffer, &size, &type);
		CreateValue((PRegistryData *)destHandle, name, type, buffer, size);
	}

	// Now take the keys
	enumHandle = source->InitKeyEnum(sourceHandle);

	while (enumHandle != NULL)
	{
		enumHandle = source->GetNextKey(enumHandle, name);

		CreateKey(destHandle, name);
		CopyKey(key + name + "/", source);
	}

	// Close the keys again
	CloseKey(destHandle);
	source->CloseKey(sourceHandle);
}
