/******************************************************************************/
/* PSkipList header file.                                                     */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#ifndef __PSkipList_h
#define __PSkipList_h

// PolyKit headers
#include "POS.h"
#include "PException.h"
#include "PList.h"
#include "GlobalFuncs.h"


/******************************************************************************/
/* PSkipList class                                                            */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
class PSkipList
{
public:
	// Constructor and destructor
	PSkipList(uint8 maxLevel = 32);
	~PSkipList(void);				// Cannot be virtual!

	// Insertion and removal of items
	void InsertItem(const KEY_TYPE key, const TYPE item);
	void RemoveItem(const KEY_TYPE key);
	TYPE GetAndRemoveItem(const KEY_TYPE key);

	// Getting and setting items
	TYPE GetItem(const KEY_TYPE key) const;
	void SetItem(const KEY_TYPE key, const TYPE newItem);
	TYPE GetAndSetItem(const KEY_TYPE key, const TYPE newItem);

	// Searching for a key
	bool HasKey(const KEY_TYPE key) const;

	// Getting keys
	void GetKeys(PList<KEY_TYPE> *list) const;

	// Emptying the list
	void MakeEmpty(void);
	bool IsEmpty(void) const;

	// Counting the items
	uint32 CountItems(void) const;

	// Assignment
	const PSkipList<TYPE, KEY_TYPE>& operator = (const PSkipList<TYPE, KEY_TYPE> &skiplist);

protected:
	class PNode;	// Predeclaration

	// PHeader helper class
	class PHeader
	{
	public:
		// Constructor
		PHeader(uint8 level)
		{
			// Initialize member variables
			numLevels = level + 1;

			// Create vector of forward node pointers
			nextNode = new PNode*[numLevels];
			if (nextNode == NULL)
				throw PMemoryException();
			
			// Set all the pointers to forward nodes to NULL
			Clean();
		}

		// Destructor deletes the next node in the list
		virtual ~PHeader(void) { delete[] nextNode; }

		// Clean() sets all the pointers to forward nodes to NULL
		void Clean(void) { for (uint32 i = 0; i < numLevels; i++) nextNode[i] = NULL; }

		// Attributes
		PNode **nextNode;	// Vector of pointers for forward nodes
		uint8 numLevels;	// Number of levels for vector of forward pointers
	};

	// PNode helper class
	class PNode : public PHeader
	{
	public:
		// Constructor
		PNode(uint8 level, KEY_TYPE key, TYPE item) : PHeader(level)
		{
			// Initialize member variables
			this->key = key;
			this->item = item;
		}

		// Attributes
		KEY_TYPE key;	// Search key
		TYPE item;		// The item
	};

	// Helper functions
	PNode* FindNode(const KEY_TYPE key) const;
	PNode* FindNodeAndUpdate(const KEY_TYPE key) const;
	PNode* RemoveNode(const KEY_TYPE key);

	// Attributes
	uint8 maxLevel;		// The max level limit
	uint8 level;		// Current level
	uint32 count;		// Number of items
	PHeader *header;	// The header for the forward nodes
	PNode **update;		// Update vector with forward node pointers
};



/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "maxLevel" is the maximal level for the list.                      */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
PSkipList<TYPE, KEY_TYPE>::PSkipList(uint8 maxLevel)
{
	// Initialize member variables
	this->maxLevel = maxLevel;
	level = 0;
	count = 0;

	// Allocate and intialize the header
	header = new PHeader(maxLevel);
	if (header == NULL)
		throw PMemoryException();

	// Allocate update vector with forward node pointers
	update = new PNode*[maxLevel];
	if (update == NULL)
	{
		delete header;				// Delete the header
		throw PMemoryException();
	}
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
PSkipList<TYPE, KEY_TYPE>::~PSkipList(void)
{
	// Make the list empty
	MakeEmpty();

	// Delete the update vector
	delete[] update;

	// Delete the header
	delete header;
}



/******************************************************************************/
/* InsertItem() inserts an item into the list.                                */
/*                                                                            */
/* Input:  "key" is the search key for the item.                              */
/*         "item" is the item to insert.                                      */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
void PSkipList<TYPE, KEY_TYPE>::InsertItem(const KEY_TYPE key, const TYPE item)
{
	// Find nearest node and update the vector with pointers to this node
	PNode *node = FindNodeAndUpdate(key);

	// Throw exception if an node already exist with the specified search key
	if ((node != NULL) && (node->key == key))
		throw PKeyException(P_KEY_ERR_ENTRY_EXIST);

	// Generate a new level for the node using the randomized probabilistic function
	uint8 newLevel = 0;
	while ((Random(0x7fffffff) < 0x3fffffff) && (newLevel < (maxLevel - 1)))
		newLevel++;

	// Update the current level, header, and update vector,
	// if the new level is higher then the current level.
	if (newLevel > level)
	{
		// Update the update vector to point at the header forward node pointers
		for (uint32 i = level + 1; i <= newLevel; i++)
			update[i] = (PNode *)header;

		// Set the current level to the new level
		level = newLevel;
	}

	// Create and initialize the new node
	node = new PNode(level, key, item);
	if (node == NULL)
		throw PMemoryException();

	// Update the new node's forward node pointers for each level, and update
	// update vector's forward node pointers to this node for each level.
	for (uint32 i = 0; i <= level; i++)
	{
		// Update the node's forward pointer
		node->nextNode[i] =	update[i]->nextNode[i];

		// Update the update vector's forward pointer to the new node
		update[i]->nextNode[i] = node;
	}

	// Increment the item counter
	count++;
}



/******************************************************************************/
/* RemoveItem() removes an item from the list.                                */
/*                                                                            */
/* Input:  "key" is the search key for the item.                              */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
void PSkipList<TYPE, KEY_TYPE>::RemoveItem(const KEY_TYPE key)
{
	// Remove the node containing the item from the list
	PNode *node = RemoveNode(key);

	// Delete the node
	delete node;

	// Adjust the current level to reflect the removal of the node
	while ((level > 0) && (header->nextNode[level] == NULL))
		level--;

	// Decrement the item counter
	count--;
}



/******************************************************************************/
/* GetAndRemoveItem() gets an item and removes it from the list.              */
/*                                                                            */
/* Input:  "key" is the search key for the item.                              */
/*                                                                            */
/* Output: the item.                                                          */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
TYPE PSkipList<TYPE, KEY_TYPE>::GetAndRemoveItem(const KEY_TYPE key)
{
	// Remove the node containing the item from the list
	PNode *node = RemoveNode(key);

	// Save the item before deleting the node
	TYPE item = node->item;

	// Delete the node
	delete node;

	// Adjust the current level to reflect the removal of the node
	while ((level > 0) && (header->nextNode[level] == NULL))
		level--;

	// Decrement the item counter
	count--;

	// Return the item
	return (item);
}



/******************************************************************************/
/* GetItem() gets an item from the list.                                      */
/*                                                                            */
/* Input:  "key" is the search key for the item.                              */
/*                                                                            */
/* Output: the item.                                                          */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
TYPE PSkipList<TYPE, KEY_TYPE>::GetItem(const KEY_TYPE key) const
{
	// Find the node with the item, and return the item
	return (FindNode(key)->item);
}



/******************************************************************************/
/* SetItem() replaces an item in the list with a new item.                    */
/*                                                                            */
/* Input:  "key" is the search key for the old item.                          */
/*         "newItem" is the new item.                                         */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
void PSkipList<TYPE, KEY_TYPE>::SetItem(const KEY_TYPE key, const TYPE newItem)
{
	// Find the node with the item, set this item to the new item
	FindNode(key)->item = newItem;
}



/******************************************************************************/
/* GetAndSetItem() gets an item in the list and replaces it with a new item.  */
/*                                                                            */
/* Input:  "key" is the search key for the item.                              */
/*         "item" is the new item.                                            */
/*                                                                            */
/* Output: the item.                                                          */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
TYPE PSkipList<TYPE, KEY_TYPE>::GetAndSetItem(const KEY_TYPE key, const TYPE newItem)
{
	// Find the node with the item
	PNode *node = FindNode(key);

	// Save the old item
	TYPE oldItem = node->item;

	// Set the item of the node to the new item
	node->item = newItem;

	// Return the old item
	return (oldItem);
}



/******************************************************************************/
/* HasKey() tests if the list has a specific key.                             */
/*                                                                            */
/* Input:  "key" is the search key.                                           */
/*                                                                            */
/* Output: true if the list has the key, or false otherwise.                  */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
bool PSkipList<TYPE, KEY_TYPE>::HasKey(const KEY_TYPE key) const
{
	// Search through all the nodes from the start of the list
	PNode *node = (PNode *)header;
	while ((node = node->nextNode[0]) != NULL)
	{
		// Return true, if the key was found
		if (node->key == key)
			return (true);
	}

	// The key could not be found
	return (false);
}



/******************************************************************************/
/* GetKeys() stores all the keys of the skip list into a list.                */
/*                                                                            */
/* Input:  "list" is the list to store the keys into.                         */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
void PSkipList<TYPE, KEY_TYPE>::GetKeys(PList<KEY_TYPE> *list) const
{
	// Make sure that the list is empty
	list->MakeEmpty();

	// Add all the node keys to the specified list
	PNode *node = (PNode *)header;
	while ((node = node->nextNode[0]) != NULL)
		list->AddTail(node->key);
}



/******************************************************************************/
/* MakeEmpty() empties the list.                                              */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
void PSkipList<TYPE, KEY_TYPE>::MakeEmpty(void)
{
	// Delete all the nodes in the list
	delete header->nextNode[0];

	// Reset the member variables
	header->Clean();
	count = 0;
	level = 0;
}



/******************************************************************************/
/* IsEmpty() tests if the list is empty.                                      */
/*                                                                            */
/* Output: true if the list is empty, or false otherwise.                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
bool PSkipList<TYPE, KEY_TYPE>::IsEmpty(void) const
{
	// Test if the bottom forward pointer of the header is NULL
	return (header->nextNode[0] == NULL ? true : false);
}



/******************************************************************************/
/* CountItems() counts and returns the number of items in the list.           */
/*                                                                            */
/* Output: the number of items in the list.                                   */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
uint32 PSkipList<TYPE, KEY_TYPE>::CountItems(void) const
{
	// Return item count
	return (count);
}



/******************************************************************************/
/* operator = (const PSkipList<TYPE, KEY_TYPE> &skiplist) assigns this list   */
/*      to the specified skip list.                                           */
/*                                                                            */
/* Input:  "skiplist" is the skip list to assign to this list.                */
/*                                                                            */
/* Output: the address of the changed skip list.                              */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
const PSkipList<TYPE, KEY_TYPE>& PSkipList<TYPE, KEY_TYPE>::operator = (const PSkipList<TYPE, KEY_TYPE> &skiplist)
{
	// Empty the list
	MakeEmpty();

	// For each node in the specified list, insert the contents of the node
	// into this list.
	PNode *next = skiplist.header->nextNode[0];
	while (next != NULL)
	{
		InsertItem(next->key, next->item);
		next = next->nextNode[0];
	}

	// Return the address of this list
	return (*this);
}



/******************************************************************************/
/* FindNode() finds a node in the list.                                       */
/*                                                                            */
/* Input:  "key" is the search key for the node.                              */
/*                                                                            */
/* Output: the pointer to the node.                                           */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
PSkipList<TYPE, KEY_TYPE>::PNode* PSkipList<TYPE, KEY_TYPE>::FindNode(const KEY_TYPE key) const
{
	PNode *node, *nextNode;

	// Trace the node topdown from the current level to the lowest level, until
	// we find the node with the key closest to, and less or equal to the search key.
	node = (PNode *)header;
	for (int32 i = level; i >= 0; i--)
	{
		// Trace the node using the forward pointers until we reach NIL or
		// or the node with the key closest to, and less or equal to the search key.
		while (((nextNode = node->nextNode[i]) != NULL) && (nextNode->key < key))
			node = nextNode;
	}
	// Set the node to the closest node
	node = node->nextNode[0];

	// Throw exception if a node could not be found with the specified search key
	if ((node == NULL) || (node->key != key))
		throw PKeyException(P_KEY_ERR_ENTRY_NOT_FOUND);

	// Return the node
	return (node);
}



/******************************************************************************/
/* FindNodeAndUpdate() finds a node in the list and updates the update vector.*/
/*                                                                            */
/* Input:  "key" is the search key for the node.                              */
/*                                                                            */
/* Output: the pointer to the node.                                           */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
PSkipList<TYPE, KEY_TYPE>::PNode* PSkipList<TYPE, KEY_TYPE>::FindNodeAndUpdate(const KEY_TYPE key) const
{
	PNode *node, *nextNode;

	// Trace the node topdown from the current level to the lowest level, until
	// we find the node with the key closest to, and less or equal to the search key.
	node = (PNode *)header;
	for (int32 i = level; i >= 0; i--)
	{
		// Trace the node using the forward pointers until we reach NIL or
		// or the node with the key closest to, and less or equal to the search key.
		while (((nextNode = node->nextNode[i]) != NULL) && (nextNode->key < key))
			node = nextNode;

		// Update the update vector's forward pointer to point at the closest node
		update[i] = node;
	}

	// Return the closest node
	return (node->nextNode[0]);
}



/******************************************************************************/
/* RemoveNode() removes a node from the list.                                 */
/*                                                                            */
/* Input:  "key" is the search key for the node.                              */
/*                                                                            */
/* Output: the pointer to the node.                                           */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
template<class TYPE, class KEY_TYPE>
PSkipList<TYPE, KEY_TYPE>::PNode* PSkipList<TYPE, KEY_TYPE>::RemoveNode(const KEY_TYPE key)
{
	// Find nearest node and update the vector with pointers to this node
	PNode *node = FindNodeAndUpdate(key);

	// Throw exception if an node does not exist with the specified search key
	if ((node == NULL) || (node->key != key))
		throw PKeyException(P_KEY_ERR_ENTRY_NOT_FOUND);

	// Update the update vector's forward pointers to point to the next node
	// from the node to remove.
	for (uint32 i = 0; i <= level; i++)
	{
		// Continue as long as the update vector's forward pointer points at
		// the node to remove.
		if (update[i]->nextNode[i] != node)
			break;

		// Set the update vector's forward pointer to point to the next node
		// from the node to remove.
		update[i]->nextNode[i] = node->nextNode[i];
	}

	// Return the node
	return (node);
}

#endif
