/******************************************************************************/
/* PSynchronize header file.                                                  */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#ifndef __PSynchronize_h
#define __PSynchronize_h

// PolyKit headers
#include "POS.h"
#include "ImportExport.h"


/******************************************************************************/
/* Error values                                                               */
/******************************************************************************/
enum PSyncError
{
	pSyncOk      = 0,
	pSyncError   = -1,
	pSyncTimeout = -2
};



/******************************************************************************/
/* Timeout values                                                             */
/******************************************************************************/
#define PSYNC_INFINITE				0xffffffff



/******************************************************************************/
/* Prototypes of global functions                                             */
/******************************************************************************/
class PSync;

#if __p_os == __p_beos && __POWERPC__
#pragma export on
#endif

_IMPEXP_PKLIB int32 AtomicIncrement(int32 *variable);
_IMPEXP_PKLIB int32 AtomicDecrement(int32 *variable);

_IMPEXP_PKLIB int32 MultipleObjectsWait(PSync **objects, int32 count, bool waitAll, uint32 timeout = PSYNC_INFINITE);



/******************************************************************************/
/* PSync class                                                                */
/******************************************************************************/
class _IMPEXP_PKLIB PSync
{
public:
	PSync(void);
	virtual ~PSync(void);

	virtual PSyncError Lock(uint32 timeout = PSYNC_INFINITE) = 0;
	virtual PSyncError Unlock(void) = 0;

#if __p_os == __p_windows

	//
	// WaitForMultibleObjects() needs to be able to get the internal handle
	// of the synchronized object.
	//
	friend int32 ::MultipleObjectsWait(PSync **objects, int32 count, bool waitAll, uint32 timeout);

protected:
	//
	// GetHandle() is used by friends to get the internal object handle
	//
	virtual HANDLE GetHandle(void) = 0;

#endif
};



/******************************************************************************/
/* PSemaphore class                                                           */
/******************************************************************************/
class _IMPEXP_PKLIB PSemaphore : public PSync
{
public:
	PSemaphore(uint32 count = 1, bool lockSameThread = false);
	virtual ~PSemaphore(void);

	virtual PSyncError Lock(uint32 timeout = PSYNC_INFINITE);
	virtual PSyncError Unlock(void);

protected:
#if __p_os == __p_beos

	sem_id semID;
	thread_id owner;
	int32 ownerCount;
	bool lockThread;

#elif __p_os == __p_windows

	virtual HANDLE GetHandle(void) { return (semaphore); }
	
	HANDLE semaphore;
	DWORD owner;
	bool lockThread;

#endif
};



/******************************************************************************/
/* PMutex class                                                               */
/******************************************************************************/
class _IMPEXP_PKLIB PMutex : public PSync
{
public:
	PMutex(bool lockSameThread = false);
	virtual ~PMutex(void);

	virtual PSyncError Lock(uint32 timeout = PSYNC_INFINITE);
	virtual PSyncError Unlock(void);

protected:
#if __p_os == __p_beos

	sem_id semID;
	thread_id owner;
	int32 ownerCount;
	bool lockThread;

#elif __p_os == __p_windows

	virtual HANDLE GetHandle(void) { return (mutex); }

	HANDLE mutex;
	DWORD owner;
	bool lockThread;

#endif
};



/******************************************************************************/
/* PEvent class                                                               */
/******************************************************************************/
class _IMPEXP_PKLIB PEvent : public PSync
{
public:
	PEvent(bool manualReset = false, bool initialState = false);
	virtual ~PEvent(void);

	virtual PSyncError Lock(uint32 timeout = PSYNC_INFINITE);
	virtual PSyncError Unlock(void);

	PSyncError SetEvent(void);
	PSyncError ResetEvent(void);

protected:
#if __p_os == __p_beos

	sem_id semID;
	bool state;
	bool manual;
	bool releaseThread;

#elif __p_os == __p_windows

	virtual HANDLE GetHandle(void) { return (event); }

	HANDLE event;

#endif
};



/******************************************************************************/
/* PSWMRLocker class                                                          */
/******************************************************************************/
class _IMPEXP_PKLIB PSWMRLocker
{
public:
	PSWMRLocker(void);
	virtual ~PSWMRLocker(void);

	PSyncError WaitToWrite(uint32 timeout = PSYNC_INFINITE);
	PSyncError DoneWriting(void);

	PSyncError WaitToRead(uint32 timeout = PSYNC_INFINITE);
	PSyncError DoneReading(void);

protected:
	PMutex *mutexNoWriter;
	PEvent *eventNoReaders;
	PSemaphore *semNumReaders;
};

#if __p_os == __p_beos && __POWERPC__
#pragma export off
#endif

#endif
